#include <proto/exec.h>
#include <proto/dos.h>
#include <proto/socket.h>
#include <proto/amisslmaster.h>
#include <proto/amissl.h>

#include <exec/exec.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <libraries/amisslmaster.h>
#include <libraries/amissl.h>
#include <clib/alib_protos.h>

#include <stdbool.h>
#include <stdint.h>

#include "Timing.h"
#include "TCPCommon.h"

const char Version[] = "$VER: AmiSSLTest 1.3 (19.10.2022) by Patrik Axelsson";

static void GenerateRandomSeed(char *buffer, int size);

static void _PrintResult(struct DosLibrary *DOSBase, struct Library *TimerBase, uint64_t totalEClocks, const char *resultFormat, ...);
#define PrintResult(totalEClocks, ...) _PrintResult(DOSBase, TimerBase, totalEClocks, __VA_ARGS__)

LONG AmiSSLTest(void) {
	struct ExecBase *SysBase = *(struct ExecBase **) 4;
	struct DosLibrary *DOSBase = NULL;
	struct RDArgs *argsResult = NULL;
	struct timerequest *timerReq = NULL;
	struct Library *SocketBase = NULL;
	struct Library *AmiSSLBase = NULL;
	struct Library *AmiSSLMasterBase = NULL;
	SSL_CTX *sslContext = NULL;
	SSL *ssl = NULL;
	long s = -1;
	
	LONG retVal = RETURN_FAIL;

	int masterBaseInited = 0;
	bool baseInited = false;
	DOSBase = (void *) OpenLibrary("dos.library", 36);
	if (NULL == DOSBase) {
		goto cleanup;
	}

	void *argsValues[8] = {0};
	argsResult = ReadArgs("HOSTNAME/A,PORT/N/A,NOOPENSSLINITSSL/S,NUMRANDSEED/N/K,RANDSEEDSIZE/N/K,NUMSSLCONNECT/N/K,REUSESSL/S,QUIET/S", (void *) argsValues, NULL);
	if (NULL == argsResult) {
		PrintFault(IoErr(), NULL);
		goto cleanup;
	}

	const char *hostname = argsValues[0];
	const long port = *(const long *) argsValues[1];
	const bool openSSLInitSSL = NULL == argsValues[2];
	const long numRandSeed = NULL != argsValues[3] ? *(long *) argsValues[3] : 1;
	const long randSeedSize = NULL != argsValues[4] ? *(long *) argsValues[4] : 128;
	const long numSSLConnect = NULL != argsValues[5] ? *(long *) argsValues[5] : 1;
	const bool reuseSSL = NULL != argsValues[6];
	const bool quiet = NULL != argsValues[7];

	timerReq = OpenTimerDevice(UNIT_VBLANK);
	if (NULL == timerReq) {
		Printf("Failed opening timer.device\n");
		goto cleanup;
	}
	struct Library *TimerBase = (void *) timerReq->tr_node.io_Device;

	SocketBase = OpenLibrary("bsdsocket.library", 3);
	if (NULL == SocketBase) {
		PutStr("Failed opening bsdsocket.library.library v3\n");
	}

	uint64_t eClocksBefore;

	eClocksBefore = GetEClocks();
	AmiSSLMasterBase = OpenLibrary("amisslmaster.library", 4);
	if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "OpenLibrary(\"%s\", %lu) = %08lx", "amisslmaster.library", 4, AmiSSLMasterBase);
	if (NULL == AmiSSLMasterBase) {
		PutStr("Failed opening amisslmaster.library v4\n");
		goto cleanup;
	}


	const unsigned amiSSLVersion = AmiSSLMasterBase->lib_Version > 4 ? AMISSL_V3xx : AMISSL_V110d;
	eClocksBefore = GetEClocks();
	masterBaseInited = InitAmiSSLMaster(amiSSLVersion, FALSE);
	if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "InitAmiSSLMaster(%02lx, %ld) = %ld", amiSSLVersion, FALSE, masterBaseInited);
	if (!masterBaseInited) {
		Printf("Failed initing AmiSSLMaster v%02lx!\n", amiSSLVersion);
		goto cleanup;
	}

	eClocksBefore = GetEClocks();
	AmiSSLBase = OpenAmiSSL();
	if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "OpenAmiSSL() = %08lx", AmiSSLBase);
	if (NULL == AmiSSLBase) {
		PutStr("Failed OpenAmiSSL()\n");
		goto cleanup;
	}

	eClocksBefore = GetEClocks();
	const int initAmiSSLRes = InitAmiSSL(
		AmiSSL_SocketBase, (Tag) SocketBase,
		TAG_DONE
	);
	if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "InitAmiSSL(%08lx, %08lx, %08lx) = %ld", AmiSSL_SocketBase, (Tag) SocketBase, TAG_DONE, initAmiSSLRes);
	baseInited = 0 == initAmiSSLRes;
	if (!baseInited) {
		PutStr("Failed InitAmiSSL()\n");
		goto cleanup;
	}

	if (openSSLInitSSL) {
		eClocksBefore = GetEClocks();
		const int openSSLInitSSLRes = OPENSSL_init_ssl(OPENSSL_INIT_SSL_DEFAULT | OPENSSL_INIT_ADD_ALL_CIPHERS | OPENSSL_INIT_ADD_ALL_DIGESTS, NULL);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "OPENSSL_init_ssl(%08lx, %lx) = %ld", OPENSSL_INIT_SSL_DEFAULT | OPENSSL_INIT_ADD_ALL_CIPHERS | OPENSSL_INIT_ADD_ALL_DIGESTS, NULL, openSSLInitSSLRes);
	}

	for (long i = 0; i < numRandSeed; i ++) {
		uint8_t rand_buffer[512];
		size_t actualRandSeedSize = randSeedSize;
		if (randSeedSize < 2) {
			actualRandSeedSize = 2;
		}
		if (randSeedSize > sizeof(rand_buffer)) {
			actualRandSeedSize = sizeof(rand_buffer);
		}
		GenerateRandomSeed(rand_buffer, actualRandSeedSize);
		eClocksBefore = GetEClocks();
		RAND_seed(rand_buffer, actualRandSeedSize);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "RAND_seed(%08lx, %lu)", rand_buffer, actualRandSeedSize);
	}

	eClocksBefore = GetEClocks();
	sslContext = SSL_CTX_new(TLS_client_method());
	if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_CTX_new(TLS_client_method()) = %08lx", sslContext);
			
	for (long i = 0; i < numSSLConnect; i++) {
		if (NULL == ssl) {
			eClocksBefore = GetEClocks();
			ssl = SSL_new(sslContext);
			if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_new(%08lx) = %08lx", sslContext, ssl);
		}

		eClocksBefore = GetEClocks();
		const int setHostnameRes = SSL_set_tlsext_host_name(ssl, hostname);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_set_tlsext_host_name(%08lx, \"%s\") = %ld", ssl, hostname, setHostnameRes);

		eClocksBefore = GetEClocks();
		s = ConnectToServer(hostname, port, 10);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "ConnectToServer(\"%s\", %ld, %ld) = %ld", hostname, port, 10, s);
		if (-1 == s) {
			goto cleanup;
		}

		eClocksBefore = GetEClocks();
		const int sslSetFdRes = SSL_set_fd(ssl, s);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_set_fd(%08lx, %ld) = %ld", ssl, s, sslSetFdRes);

		eClocksBefore = GetEClocks();
		const int sslConnectRes = SSL_connect(ssl);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_connect(%08lx) = %ld", ssl, sslConnectRes);
		if (sslConnectRes <= 0) {
			Printf("Failed SSL_connect: %ld\n", SSL_get_error(ssl, sslConnectRes));
			goto cleanup;
		}

		eClocksBefore = GetEClocks();
		const char *sslVersion = SSL_get_version(ssl);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_get_version(%08lx) = \"%s\"", ssl, sslVersion);
		
		eClocksBefore = GetEClocks();
		const char *sslCipher = SSL_get_cipher(ssl);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_get_cipher(%08lx) = \"%s\"", ssl, sslCipher);

		eClocksBefore = GetEClocks();
		const int sslShutdownRes = SSL_shutdown(ssl);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_shutdown(%08lx) = %ld", ssl, sslShutdownRes);
		if (sslShutdownRes < 0) {
			Printf("Failed SSL_shutdown: %ld\n", SSL_get_error(ssl, sslShutdownRes));
			goto cleanup;
		}
		CloseSocket(s);
		s = -1;

		if (!reuseSSL) {
			eClocksBefore = GetEClocks();
			SSL_free(ssl);
			if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_free(%08lx)", ssl);
			ssl = NULL;
		}
	}

	retVal = RETURN_OK;

cleanup:
	if (s >= 0) {
		CloseSocket(s);
	}
	if (NULL != ssl) {
		eClocksBefore = GetEClocks();
		SSL_free(ssl);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_free(%08lx)", ssl);
	}
	if (NULL != sslContext) {
		eClocksBefore = GetEClocks();
		SSL_CTX_free(sslContext);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "SSL_CTX_free(%08lx)", sslContext);
	}
	if (baseInited) {
		eClocksBefore = GetEClocks();
		CleanupAmiSSL(TAG_DONE);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "CleanupAmiSSL(%08lx)", TAG_DONE);
	}
	if (NULL != AmiSSLBase) {
		eClocksBefore = GetEClocks();
		CloseAmiSSL();
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "CloseAmiSSL()");
	}
	if (NULL != AmiSSLMasterBase) {
		eClocksBefore = GetEClocks();
		CloseLibrary(AmiSSLMasterBase);
		if (!quiet) PrintResult(GetEClocks() - eClocksBefore, "CloseLibrary(%08lx)", AmiSSLMasterBase);
	}
	CloseLibrary(SocketBase);
	CloseTimerDevice(timerReq);
	if (NULL != argsResult) {
		FreeArgs(argsResult);
	}
	CloseLibrary((void *) DOSBase);

	return retVal;
}

static void GenerateRandomSeed(char *buffer, int size)
{
	int i;
	for(i = 0; i < size/2; i++)
	{
		((UWORD *)buffer)[i] = RangeRand(65535);
	}
}

static void _PrintResult(struct DosLibrary *DOSBase, struct Library *TimerBase, uint64_t totalEClocks, const char *resultFormat, ...) {
	va_list varArgs;
	va_start(varArgs, resultFormat);
	const int count = VPrintf(resultFormat, varArgs);
	va_end(varArgs);
	// This one controls how much the formatted resultFormat will be padded on
	// the right side.
	const char *spaces = "                                                               ";
	const size_t numSpaces = strlen(spaces);
	if (count < numSpaces) {
		PutStr(spaces + count);
	}
	int spacesToAdd = numSpaces - count;
	struct timeval duration = EClocksToTimeval(totalEClocks);
	Printf("%lu.%06lus\n", duration.tv_secs, duration.tv_micro);
}
