#include <exec/types.h>
#include <dos/dos.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <proto/dos.h>
#include <proto/socket.h>

#include <stdbool.h>
#include <string.h>

static ULONG GetTagData(
		struct Library *SocketBase,
		const LONG tag,
		const long inputData
) {
	struct TagItem tags[] = {
		{ .ti_Tag = tag, .ti_Data = inputData },
		{ .ti_Tag = TAG_END }};
	if (0 == SocketBaseTagList(tags)) {
		return tags[0].ti_Data;
	}
	return 0;
}

static void PrintError(
		struct DosLibrary *DOSBase,
		struct Library *SocketBase,
		const char *header,
		const long errnoNum,
		const ULONG errnoStringTag
) {
	if (0 != errnoNum) {
		const char *errnoString = (void *) GetTagData(SocketBase, errnoStringTag, errnoNum);
		Printf("%s%s%s (%ld)\n", NULL != header ? header : "", NULL != header ? ": " : "", errnoString, errnoNum);
	}
}

static long GetSocketError(struct Library *SocketBase, const long socketDesc) {
	long socketError = 0;
	socklen_t socketErrorLen = sizeof(socketError);
	getsockopt(socketDesc, SOL_SOCKET, SO_ERROR, &socketError, &socketErrorLen);
	return socketError;
}

void _PrintSocketError(
		struct DosLibrary *DOSBase,
		struct Library *SocketBase,
		const char *header
) {
	PrintError(DOSBase, SocketBase, header, GetTagData(SocketBase, SBTM_GETVAL(SBTC_ERRNO), 0), SBTM_GETVAL(SBTC_ERRNOSTRPTR));
}

void PrintResolverError(
		struct DosLibrary *DOSBase,
		struct Library *SocketBase,
		const char *header
) {
	PrintError(DOSBase, SocketBase, header, GetTagData(SocketBase, SBTM_GETVAL(SBTC_HERRNO), 0), SBTM_GETVAL(SBTC_HERRNOSTRPTR));
} 

void PrintSocketSocketError(
		struct DosLibrary *DOSBase,
		struct Library *SocketBase,
		const char *header,
		const long errnoNum
) {
	PrintError(DOSBase, SocketBase, header, errnoNum, SBTM_GETVAL(SBTC_ERRNOSTRPTR));
} 

long _ConnectToServer(
		struct DosLibrary *DOSBase,
		struct Library *SocketBase,
		const char *serverHostname,
		const short serverPort,
		const long timeoutSeconds
) {
	struct hostent *serverHostent = gethostbyname((char *) serverHostname);
	if(NULL == serverHostent) {
		PrintResolverError(DOSBase, SocketBase, serverHostname);
		return -1;
	}
	struct sockaddr_in serverAddr = {
		.sin_family = AF_INET,
		.sin_port = htons(serverPort),
		//.sin_addr.s_addr = *(unsigned int *) serverHostent->h_addr_list[0]
	};
	// Handle this one with a separate memcpy as certain stacks (AmiTCP)
	// returns resolver addresses at odd addresses. The above solution
	// with cast to int would give a #80000003 on a 68000 or 68010 as they
	// cannot handle longword access at odd addresses.
	memcpy(&serverAddr.sin_addr.s_addr, serverHostent->h_addr_list[0], sizeof(serverAddr.sin_addr.s_addr));
	
	long socketDesc = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP);
	if(-1 == socketDesc) {
		_PrintSocketError(DOSBase, SocketBase, "socket()");
		goto fail;
	}

	// The reason for attempting to set the socket in non-blocking mode is to be
	// able to do a connect with a timeout lower than whatever the TCP/IP-stack
	// has dictated as suitable (usually 1 - 2minutes if the host isn't responding).
	// Unfortunately, some versions of UAE, like uae4arm claims they have
	// bsdsocket.library v3 compatibility, but they still cannot cannot handle
	// for example non-blocking sockets. So instead of just failing here, do a
	// regular blocking connect if the socket cannot be set to non-blocking mode.
	if(timeoutSeconds > 0 &&
	   -1 != IoctlSocket(socketDesc, FIONBIO, (char *) &(int) {1})) {
		if(-1 == connect(socketDesc, (struct sockaddr*) &serverAddr, sizeof(serverAddr))) {
			const long socketError = GetTagData(SocketBase, SBTM_GETVAL(SBTC_ERRNO), 0);
			if(EINPROGRESS != socketError) {
				PrintSocketSocketError(DOSBase, SocketBase, serverHostname, socketError);
				goto fail;
			}
		}

		fd_set fds;
		FD_ZERO(&fds);
		FD_SET(socketDesc, &fds);
		const long selectResult = WaitSelect(socketDesc + 1, NULL, &fds, NULL, &(struct timeval) { .tv_sec = timeoutSeconds }, NULL);
		if(-1 == selectResult) {
			_PrintSocketError(DOSBase, SocketBase, "WaitSelect()");
			goto fail;
		}
		else if(0 == selectResult) {
			Printf("%s: Timed out connecting\n", serverHostname);
			goto fail;
		}
		else {
			// As we are doing non-blocking I/O, we need to get potential
			// errors separately
			const long socketError = GetSocketError(SocketBase, socketDesc);
			if(0 != socketError) {
				PrintSocketSocketError(DOSBase, SocketBase, serverHostname, socketError);
				goto fail;
			}
		}

		const long disableValue = 0;
		if(-1 == IoctlSocket(socketDesc, FIONBIO, (char *) &disableValue)) {
			_PrintSocketError(DOSBase, SocketBase, "IoctlSocket()");
			goto fail;
		}
	}
	else {
		// Regular blocking connect
		if(-1 == connect(socketDesc, (struct sockaddr*) &serverAddr, sizeof(serverAddr))) {
			_PrintSocketError(DOSBase, SocketBase, serverHostname);
			goto fail;
		}
	}

	return socketDesc;

fail:
	CloseSocket(socketDesc);
	return -1;
}
